
/*

    Goldleaf - Multipurpose homebrew tool for Nintendo Switch
    Copyright (C) 2018-2023 XorTroll

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

*/

#include <ui/ui_ContentInformationLayout.hpp>
#include <ui/ui_MainApplication.hpp>

extern ui::MainApplication::Ref g_MainApplication;
extern cfg::Settings g_Settings;

namespace {

    inline std::string FormatPlayStatsInfo(const std::string &kind, const hos::TitlePlayStats &stats) {
        std::string fmt;
        fmt += "\n" + kind;
        fmt += "\n" + cfg::Strings.GetString(339) + " " + hos::FormatTime(stats.secs_from_last_launched);
        fmt += "\n" + cfg::Strings.GetString(340) + " " + hos::FormatTime(stats.total_play_secs);
        return fmt;
    }

}

namespace ui {

    ContentInformationLayout::ContentInformationLayout() {
        this->options_menu = pu::ui::elm::Menu::New(0, 160, pu::ui::render::ScreenWidth, g_Settings.custom_scheme.base, g_Settings.custom_scheme.base_focus, g_Settings.menu_item_size, ComputeDefaultMenuItemCount(g_Settings.menu_item_size));
        g_Settings.ApplyScrollBarColor(this->options_menu);
        this->Add(this->options_menu);
    }

    void ContentInformationLayout::UpdateElements() {
        this->options_menu->ClearItems();
        for(const auto &sub_cnt: this->cnt_subcontents) {
            auto name = cfg::Strings.GetString(261);
            if(sub_cnt.IsUpdate()) {
                name = cfg::Strings.GetString(262);
            }
            if(sub_cnt.IsAddOnContent()) {
                name = cfg::Strings.GetString(263) + " " + std::to_string(hos::GetIdFromAddOnContentApplicationId(sub_cnt.app_id));
            }
            auto sub_cnt_item = pu::ui::elm::MenuItem::New(name);
            sub_cnt_item->SetColor(g_Settings.custom_scheme.text);
            sub_cnt_item->AddOnKey(std::bind(&ContentInformationLayout::options_DefaultKey, this));
            this->options_menu->AddItem(sub_cnt_item);
        }
        this->options_menu->SetSelectedIndex(0);
    }

    void ContentInformationLayout::options_DefaultKey() {
        const auto idx = this->options_menu->GetSelectedIndex();
        auto msg = cfg::Strings.GetString(169) + "\n\n";
        msg += cfg::Strings.GetString(170) + " ";
        std::string icon;
        auto &sub_cnt = this->cnt_subcontents.at(idx);
        auto sd_exp = fs::GetSdCardExplorer();
        if(sd_exp->IsFile(hos::GetExportedIconPath(sub_cnt.app_id))) {
            icon = hos::GetExportedIconPath(sub_cnt.app_id);
        }
        switch(sub_cnt.type) {
            case NcmContentMetaType_Application: {
                msg += cfg::Strings.GetString(171);
                break;
            }
            case NcmContentMetaType_AddOnContent: {
                msg += cfg::Strings.GetString(172);
                break;
            }
            case NcmContentMetaType_Patch: {
                msg += cfg::Strings.GetString(173);
                break;
            }
            case NcmContentMetaType_SystemProgram: {
                msg += cfg::Strings.GetString(174);
                break;
            }
            case NcmContentMetaType_SystemData: {
                msg += cfg::Strings.GetString(175);
                break;
            }
            default: {
                msg += cfg::Strings.GetString(176);
                break;
            }
        }
        msg += "\n" + cfg::Strings.GetString(90) + " " + hos::FormatApplicationId(sub_cnt.app_id);
        size_t total_size = 0;
        if(sub_cnt.TryGetContents()) {
            for(const auto &cnts: sub_cnt.title_cnts) {
                total_size += cnts.GetTotalSize();        
            }
        }
        
        if(sub_cnt.storage_id == NcmStorageId_None) {
            msg += "\n\n" + cfg::Strings.GetString(463);
        }
        else {
            msg += "\n\n" + cfg::Strings.GetString(177) + " " + fs::FormatSize(total_size);
            msg += "\n\n" + cfg::Strings.GetString(178) + " v" + std::to_string(sub_cnt.version);
        }
        if(sub_cnt.version != 0) {
            msg += " [" + cfg::Strings.GetString(179) + " no. " + std::to_string(sub_cnt.version >> 16) + "]";
        }

        if(sub_cnt.IsBaseTitle() && (sub_cnt.storage_id != NcmStorageId_BuiltInSystem)) {
            msg += "\n";

            const auto global_stats = sub_cnt.GetGlobalPlayStats();
            if(global_stats.total_play_secs == 0) {
                msg += "\n" + cfg::Strings.GetString(351) + "\n";
            }
            else {
                msg += FormatPlayStatsInfo(cfg::Strings.GetString(338), global_stats);
                if(acc::HasSelectedUser()) {
                    const auto user_stats = sub_cnt.GetUserPlayStats(acc::GetSelectedUser());
                    msg += "\n";
                    msg += FormatPlayStatsInfo(cfg::Strings.GetString(337), user_stats);
                }
            }
        }
        const auto tiks = hos::GetAllTickets();
        const auto it = std::find_if(tiks.begin(), tiks.end(), [&](const hos::Ticket &tik) {
            return tik.rights_id.GetApplicationId() == sub_cnt.app_id;
        });

        const bool has_tik = it != tiks.end();
        hos::Ticket tik = {};
        if(has_tik) {
            tik = *it;
        }

        std::vector<std::string> options;
        int remove_title_opt = INT32_MIN;
        int export_opt = INT32_MIN;
        int mount_save_data_opt = INT32_MIN;
        int remove_tik_opt = INT32_MIN;
        int reset_launch_ver_opt = INT32_MIN;
        u32 cur_last_opt = 0;

        if(sub_cnt.storage_id != NcmStorageId_GameCard) {
            if(sub_cnt.storage_id != NcmStorageId_BuiltInSystem) {
                options.push_back(cfg::Strings.GetString(245));
                remove_title_opt = cur_last_opt; cur_last_opt++;
            }

            if(sub_cnt.storage_id != NcmStorageId_None) {
                options.push_back(cfg::Strings.GetString(244));
                export_opt = cur_last_opt; cur_last_opt++;
            }
        }

        if(sub_cnt.storage_id != NcmStorageId_BuiltInSystem) {
            if(sub_cnt.IsBaseTitle()) {
                options.push_back(cfg::Strings.GetString(414));
                mount_save_data_opt = cur_last_opt; cur_last_opt++;
            }

            if(sub_cnt.storage_id != NcmStorageId_None) {
                if(has_tik) {
                    options.push_back(cfg::Strings.GetString(293));
                    remove_tik_opt = cur_last_opt; cur_last_opt++;
                }
            }

            options.push_back(cfg::Strings.GetString(319));
            reset_launch_ver_opt = cur_last_opt; cur_last_opt++;
        }

        options.push_back(cfg::Strings.GetString(18));

        const auto option_1 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(243), msg, options, true, icon);
        if(option_1 == remove_title_opt) {
            const auto option_2 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(243), cfg::Strings.GetString(186), { cfg::Strings.GetString(111), cfg::Strings.GetString(18) }, true);
            if(option_2 == 0) {
                auto remove_tik = false;
                if(has_tik) {
                    const auto option_3 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(243), cfg::Strings.GetString(204), { cfg::Strings.GetString(111), cfg::Strings.GetString(112), cfg::Strings.GetString(18) }, true);
                    remove_tik = option_3 == 0;
                }

                hos::RemoveTitle(sub_cnt);
                if(remove_tik) {
                    hos::RemoveTicket(tik);
                }
                g_MainApplication->ShowNotification(cfg::Strings.GetString(246));
                this->cnt_subcontents.erase(this->cnt_subcontents.begin() + idx);
                if(this->cnt_subcontents.empty()) {
                    g_MainApplication->GetStorageContentsLayout()->Reload();
                    g_MainApplication->LoadMenuData(cfg::Strings.GetString(32), "Storage", cfg::Strings.GetString(33));
                    g_MainApplication->LoadLayout(g_MainApplication->GetStorageContentsLayout());
                }
                else {
                    this->UpdateElements();
                }
            }
        }
        else if(option_1 == export_opt) {
            const auto option_2 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(182), cfg::Strings.GetString(184), { cfg::Strings.GetString(111), cfg::Strings.GetString(18) }, true);
            if(option_2 == 0) {
                g_MainApplication->LoadLayout(g_MainApplication->GetContentExportLayout());
                g_MainApplication->GetContentExportLayout()->StartExport(sub_cnt, has_tik);
                g_MainApplication->LoadLayout(g_MainApplication->GetContentInformationLayout());
            }
        }
        else if(option_1 == mount_save_data_opt) {
            if(!acc::HasSelectedUser()) {
                g_MainApplication->CreateShowDialog(cfg::Strings.GetString(408), cfg::Strings.GetString(411), { cfg::Strings.GetString(234) }, true);
                return;
            }

            FsFileSystem save_data_fs;
            const auto rc = fs::MountTitleSaveData(sub_cnt.app_id, acc::GetSelectedUser(), save_data_fs);
            if(R_SUCCEEDED(rc)) {
                auto save_data_exp = new fs::FspExplorer(save_data_fs, "fs.SaveData-0x" + hos::FormatApplicationId(sub_cnt.app_id));
                save_data_exp->SetShouldWarnOnWriteAccess(true);
                g_MainApplication->GetExploreMenuLayout()->AddMountedExplorer(save_data_exp, cfg::Strings.GetString(408), icon);
                g_MainApplication->ShowNotification(cfg::Strings.GetString(412));
            }
            else {
                HandleResult(rc, cfg::Strings.GetString(413));
            }
        }
        else if(option_1 == remove_tik_opt) {
            const auto option_2 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(200), cfg::Strings.GetString(205), { cfg::Strings.GetString(111), cfg::Strings.GetString(18) }, true);
            if(option_2 == 0) {
                const auto rc = es::DeleteTicket(tik.rights_id);
                if(R_SUCCEEDED(rc)) {
                    g_MainApplication->ShowNotification(cfg::Strings.GetString(206));
                    this->UpdateElements();
                }
                else {
                    HandleResult(rc, cfg::Strings.GetString(207));
                }
            }
        }
        else if(option_1 == reset_launch_ver_opt) {
            const auto option_2 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(319), cfg::Strings.GetString(457), { cfg::Strings.GetString(111), cfg::Strings.GetString(18) }, true);
            if(option_2 == 0) {
                const auto rc = hos::UpdateTitleVersion(sub_cnt);
                if(R_SUCCEEDED(rc)) {
                    g_MainApplication->ShowNotification(cfg::Strings.GetString(322));
                    this->UpdateElements();
                }
                else {
                    HandleResult(rc, cfg::Strings.GetString(234));
                }
            }
        }
    }

    void ContentInformationLayout::LoadContent(hos::Title &content) {
        this->cnt_subcontents.clear();
        this->cnt_subcontents.push_back(content);
        const auto &all_contents = hos::SearchTitles(content.storage_id, NcmContentMetaType_Unknown);
        for(const auto &cnt: all_contents) {
            if(content.IsBaseOf(cnt)) {
                this->cnt_subcontents.push_back(cnt);
            }
        }

        auto head = hos::FormatApplicationId(content.app_id);
        if(content.TryGetNacp()) {
            head = hos::FindNacpName(content.nacp) + " (" + content.nacp.display_version + ")";
        }

        g_MainApplication->LoadMenuData(cfg::Strings.GetString(187), hos::GetExportedIconPath(content.app_id), head, false);
        this->UpdateElements();
    }

}