
/*

    Goldleaf - Multipurpose homebrew tool for Nintendo Switch
    Copyright (C) 2018-2023 XorTroll

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

*/

#include <ui/ui_PartitionBrowserLayout.hpp>
#include <ui/ui_MainApplication.hpp>
#include <hos/hos_Payload.hpp>
#include <amssu/amssu_Service.hpp>

extern ui::MainApplication::Ref g_MainApplication;
extern cfg::Settings g_Settings;

namespace ui {

    namespace {

        std::stack<u32> g_EntryIndexStack;

        inline bool IsHiddenContent(const std::string &path) {
            return !path.empty() && (path[0] == '.');
        }

    }

    PartitionBrowserLayout::PartitionBrowserLayout() : pu::ui::Layout() {
        this->cur_exp = fs::GetSdCardExplorer();
        this->browse_menu = pu::ui::elm::Menu::New(0, 160, pu::ui::render::ScreenWidth, g_Settings.custom_scheme.base, g_Settings.custom_scheme.base_focus, g_Settings.menu_item_size, ComputeDefaultMenuItemCount(g_Settings.menu_item_size));
        g_Settings.ApplyScrollBarColor(this->browse_menu);
        this->empty_dir_text = pu::ui::elm::TextBlock::New(30, 630, cfg::Strings.GetString(49));
        this->empty_dir_text->SetHorizontalAlign(pu::ui::elm::HorizontalAlign::Center);
        this->empty_dir_text->SetVerticalAlign(pu::ui::elm::VerticalAlign::Center);
        this->empty_dir_text->SetColor(g_Settings.custom_scheme.text);
        this->Add(this->browse_menu);
        this->Add(this->empty_dir_text);
    }

    void PartitionBrowserLayout::ChangePartitionExplorer(fs::Explorer *exp, const bool update_contents) {
        this->cur_exp = exp;
        if(update_contents) {
            this->UpdateElements();
        }
    }

    void PartitionBrowserLayout::ChangePartitionSdCard(const bool update_contents) {
        this->ChangePartitionExplorer(fs::GetSdCardExplorer(), update_contents);
    }

    void PartitionBrowserLayout::ChangePartitionNAND(const fs::Partition partition, const bool update_contents) {
        fs::Explorer *exp = nullptr;
        switch(partition) {
            case fs::Partition::PRODINFOF: {
                exp = fs::GetPRODINFOFExplorer();
                break;
            }
            case fs::Partition::NANDSafe: {
                exp = fs::GetNANDSafeExplorer();
                break;
            }
            case fs::Partition::NANDUser: {
                exp = fs::GetNANDUserExplorer();
                break;
            }
            case fs::Partition::NANDSystem: {
                exp = fs::GetNANDSystemExplorer();
                break;
            }
            default: {
                return;
            }
        }
        this->ChangePartitionExplorer(exp, update_contents);
    }
    
    void PartitionBrowserLayout::ChangePartitionPCDrive(const std::string &mount_name, const bool update_contents) {
        this->ChangePartitionExplorer(fs::GetRemotePCExplorer(mount_name), update_contents);
    }

    void PartitionBrowserLayout::ChangePartitionDrive(UsbHsFsDevice &drv, const bool update_contents) {
        this->ChangePartitionExplorer(fs::GetDriveExplorer(drv), update_contents);
    }

    void PartitionBrowserLayout::UpdateElements(const int idx) {
        const auto contents = this->cur_exp->GetContents();
        this->browse_menu->ClearItems();
        this->ResetMenuHead();
        this->browse_menu->SetVisible(!contents.empty());
        this->empty_dir_text->SetVisible(contents.empty());
        if(!contents.empty()) {
            for(const auto &item: contents) {
                if(g_Settings.ignore_hidden_files && IsHiddenContent(item)) {
                    continue;
                }

                auto menu_item = pu::ui::elm::MenuItem::New(item);
                menu_item->SetColor(g_Settings.custom_scheme.text);
                if(this->cur_exp->IsDirectory(item)) {
                    menu_item->SetIcon(g_Settings.PathForResource("/FileSystem/Directory.png"));
                }
                else {
                    const auto ext = LowerCaseString(fs::GetExtension(item));
                    if(ext == "nsp") {
                        menu_item->SetIcon(g_Settings.PathForResource("/FileSystem/NSP.png"));
                    }
                    else if(ext == "nro") {
                        menu_item->SetIcon(g_Settings.PathForResource("/FileSystem/NRO.png"));
                    }
                    else if(ext == "tik") {
                        menu_item->SetIcon(g_Settings.PathForResource("/FileSystem/TIK.png"));
                    }
                    else if(ext == "cert") {
                        menu_item->SetIcon(g_Settings.PathForResource("/FileSystem/CERT.png"));
                    }
                    /*
                    else if(ext == "nxtheme") {
                        menu_item->SetIcon(g_Settings.PathForResource("/FileSystem/NXTheme.png"));
                    }
                    */
                    else if(ext == "nca") {
                        menu_item->SetIcon(g_Settings.PathForResource("/FileSystem/NCA.png"));
                    }
                    else if(ext == "nacp") {
                        menu_item->SetIcon(g_Settings.PathForResource("/FileSystem/NACP.png"));
                    }
                    else if((ext == "jpg") || (ext == "jpeg")) {
                        menu_item->SetIcon(g_Settings.PathForResource("/FileSystem/JPEG.png"));
                    }
                    else {
                        menu_item->SetIcon(g_Settings.PathForResource("/FileSystem/File.png"));
                    }
                }
                menu_item->AddOnKey(std::bind(&PartitionBrowserLayout::fsItems_DefaultKey, this, item));
                menu_item->AddOnKey(std::bind(&PartitionBrowserLayout::fsItems_Y, this, item), HidNpadButton_Y);
                this->browse_menu->AddItem(menu_item);
            }
            u32 tmp_idx = 0;
            if(idx < 0) {
                if(!g_EntryIndexStack.empty()) {
                    tmp_idx = g_EntryIndexStack.top();
                    g_EntryIndexStack.pop();
                }
            }
            else {
                tmp_idx = static_cast<u32>(idx);
                if(tmp_idx >= contents.size()) {
                    tmp_idx = 0;
                }
            }
            this->browse_menu->SetSelectedIndex(tmp_idx);
        }
    }

    void PartitionBrowserLayout::ResetMenuHead() {
        g_MainApplication->LoadMenuHead(this->cur_exp->GetPresentableCwd());
    }

    void PartitionBrowserLayout::HandleFileDirectly(const std::string &path) {
        const auto dir = fs::GetBaseDirectory(path);
        const auto file_name = fs::GetBaseName(path);
        this->ChangePartitionPCDrive(dir);

        auto &items = this->browse_menu->GetItems();
        const auto it = std::find_if(items.begin(), items.end(), [&](pu::ui::elm::MenuItem::Ref &item) -> bool {
            return item->GetName() == file_name;
        });
        if(it == items.end()) {
            return;
        }

        const auto idx = std::distance(items.begin(), it);
        this->browse_menu->SetSelectedIndex(idx);
        this->fsItems_DefaultKey(file_name);
    }

    void PartitionBrowserLayout::PromptDeleteFile(const std::string &path) {
        if(this->CheckWriteAccess()) {
            const auto option = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(127), cfg::Strings.GetString(128), { cfg::Strings.GetString(111), cfg::Strings.GetString(18) }, true);
            if(option < 0) {
                return;
            }

            this->cur_exp->DeleteFile(path);
            g_MainApplication->ShowNotification(cfg::Strings.GetString(129));
            auto tmp_idx = this->browse_menu->GetSelectedIndex();
            if(tmp_idx > 0) {
                tmp_idx--;
            }
            this->UpdateElements(tmp_idx);
        }
    }

    bool PartitionBrowserLayout::GoBack() {
        return this->cur_exp->NavigateBack();
    }

    bool PartitionBrowserLayout::CheckWriteAccess() {
        if(this->cur_exp->IsWriteAccessBlocked()) {
            g_MainApplication->ShowNotification(cfg::Strings.GetString(476));
            return false;
        }
        else if(this->cur_exp->ShouldWarnOnWriteAccess()) {
            const auto option = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(50), cfg::Strings.GetString(51), { cfg::Strings.GetString(111), cfg::Strings.GetString(18) }, true);
            return option == 0;
        }
        else {
            return true;
        }
    }

    void PartitionBrowserLayout::fsItems_DefaultKey(const std::string &item) {
        const auto full_item = this->cur_exp->FullPathFor(item);
        const auto pres_full_item = this->cur_exp->FullPresentablePathFor(item);
        if(this->cur_exp->NavigateForward(full_item)) {
            g_EntryIndexStack.push(this->browse_menu->GetSelectedIndex());
            this->UpdateElements();
        }
        else {
            const auto ext = LowerCaseString(fs::GetExtension(item));
            std::string icon_path = "";
            auto msg = cfg::Strings.GetString(52) + " ";
            if(ext == "nsp") {
                msg += cfg::Strings.GetString(53);
            }
            else if(ext == "nro") {
                msg += cfg::Strings.GetString(54);
            }
            else if(ext == "tik") {
                msg += cfg::Strings.GetString(55);
            }
            /*
            else if(ext == "nxtheme") {
                msg += cfg::Strings.GetString(56);
            }
            */
            else if(ext == "nca") {
                msg += cfg::Strings.GetString(57);
            }
            else if(ext == "nacp") {
                msg += cfg::Strings.GetString(58);
            }
            else if((ext == "jpg") || (ext == "jpeg")) {
                msg += cfg::Strings.GetString(59);
                icon_path = full_item;
            }
            else if(ext == "png") {
                icon_path = full_item;
            }
            else {
                msg += cfg::Strings.GetString(270);
            }
            msg += "\n\n" + cfg::Strings.GetString(64) + " " + fs::FormatSize(this->cur_exp->GetFileSize(full_item));
            const auto is_bin = this->cur_exp->IsFileBinary(full_item);
            std::vector<std::string> dialog_opts;
            u32 option_count = 5;
            if(ext == "nsp") {
                dialog_opts.push_back(cfg::Strings.GetString(65));
                option_count++;
            }
            else if(ext == "nro") {
                dialog_opts.push_back(cfg::Strings.GetString(66));
                option_count++;
                dialog_opts.push_back(cfg::Strings.GetString(453));
                option_count++;
            }
            else if(ext == "tik") {
                dialog_opts.push_back(cfg::Strings.GetString(67));
                option_count++;
            }
            /*
            else if(ext == "nxtheme") {
                dialog_opts.push_back(cfg::Strings.GetString(65));
                option_count++;
            }
            */
            else if(ext == "nacp") {
                dialog_opts.push_back(cfg::Strings.GetString(69));
                option_count++;
            }
            else if((ext == "jpg") || (ext == "jpeg")) {
                dialog_opts.push_back(cfg::Strings.GetString(70));
                option_count++;
            }
            else if(ext == "bin") {
                dialog_opts.push_back(cfg::Strings.GetString(66));
                option_count++;
            }
            else if(!is_bin) {
                dialog_opts.push_back(cfg::Strings.GetString(71));
                option_count++;
            }
            dialog_opts.push_back(cfg::Strings.GetString(72));
            dialog_opts.push_back(cfg::Strings.GetString(73));
            dialog_opts.push_back(cfg::Strings.GetString(74));
            dialog_opts.push_back(cfg::Strings.GetString(75));
            dialog_opts.push_back(cfg::Strings.GetString(18));
            const auto option_1 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(76), msg, dialog_opts, true, icon_path);
            if(option_1 < 0) {
                return;
            }

            if(ext == "nsp") {
                switch(option_1) {
                    case 0: {
                        const auto option_2 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(77), cfg::Strings.GetString(78), { cfg::Strings.GetString(19), cfg::Strings.GetString(79), cfg::Strings.GetString(18) }, true);
                        if(option_2 < 0) {
                            return;
                        }
                        const auto dst = (option_2 == 0) ? NcmStorageId_SdCard : NcmStorageId_BuiltInUser;
                        const auto file_size = this->cur_exp->GetFileSize(full_item);
                        const auto free_space = fs::GetFreeSpaceForPartition(static_cast<fs::Partition>(dst));
                        if(free_space < file_size) {
                            HandleResult(rc::goldleaf::ResultNotEnoughSize, cfg::Strings.GetString(251));
                            return;
                        }
                        g_MainApplication->LoadMenuHead(cfg::Strings.GetString(145) + " " + pres_full_item);
                        g_MainApplication->LoadLayout(g_MainApplication->GetInstallLayout());
                        g_MainApplication->GetInstallLayout()->StartInstall(full_item, this->cur_exp, dst);
                        g_MainApplication->LoadLayout(g_MainApplication->GetBrowserLayout());
                        this->ResetMenuHead();
                        break;
                    }
                }
            }
            else if(ext == "nro") {
                switch(option_1) {
                    case 0: {
                        if(GetExecutableMode() == ExecutableMode::NRO) {
                            const auto option_2 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(98), cfg::Strings.GetString(99), { cfg::Strings.GetString(66), cfg::Strings.GetString(18) }, true);
                            if(option_2 < 0) {
                                return;
                            }
                            envSetNextLoad(full_item.c_str(), full_item.c_str());
                            g_MainApplication->CloseWithFadeOut();
                            return;
                        }
                        else {
                            g_MainApplication->CreateShowDialog(cfg::Strings.GetString(98), cfg::Strings.GetString(100), { cfg::Strings.GetString(234) }, false);
                            return;
                        }
                        break;
                    }
                    case 1: {
                        NroHeader nro_h = {};
                        size_t romfs_offset = 0;
                        this->cur_exp->StartFile(full_item, fs::FileMode::Read);
                        if(this->cur_exp->ReadFile(full_item, sizeof(NroStart), sizeof(nro_h), &nro_h) == sizeof(nro_h)) {
                            if(nro_h.magic == NROHEADER_MAGIC) {
                                NroAssetHeader nro_asset_h = {};
                                if(this->cur_exp->ReadFile(full_item, nro_h.size, sizeof(nro_asset_h), &nro_asset_h) == sizeof(nro_asset_h)) {
                                    if(nro_asset_h.magic == NROASSETHEADER_MAGIC) {
                                        if(nro_asset_h.romfs.size > 0) {
                                            romfs_offset = nro_h.size + nro_asset_h.romfs.offset;
                                        }
                                    }
                                }
                            }
                        }
                        this->cur_exp->EndFile();

                        if(romfs_offset > 0) {
                            const auto romfs_id = std::to_string(randomGet64());
                            const auto mnt_name = "nroromfs-" + romfs_id;
                            const auto rc = romfsMountFromFsdev(full_item.c_str(), romfs_offset, mnt_name.c_str());
                            if(R_SUCCEEDED(rc)) {
                                auto romfs_exp = new fs::RomFsExplorer("NroRomFs-" + romfs_id, mnt_name, true);
                                romfs_exp->SetWriteAccessBlocked(true);
                                g_MainApplication->GetExploreMenuLayout()->AddMountedExplorer(romfs_exp, "RomFs (" + fs::GetBaseName(full_item) + ")", g_Settings.PathForResource("/Common/NAND.png"));
                                g_MainApplication->ShowNotification(cfg::Strings.GetString(454));
                            }
                            else {
                                romfs_offset = 0;
                            }
                        }
                        
                        if(romfs_offset == 0) {
                            g_MainApplication->ShowNotification(cfg::Strings.GetString(455));
                        }
                        break;
                    }
                }
            }
            else if(ext == "tik") {
                switch(option_1) {
                    case 0: {
                        const auto option_2 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(101), cfg::Strings.GetString(102), { cfg::Strings.GetString(234), cfg::Strings.GetString(18) }, true);
                        if(option_2 == 0) {
                            const auto tik_file_size = this->cur_exp->GetFileSize(full_item);
                            auto tik_read_buf = fs::AllocateWorkBuffer(tik_file_size);
                            this->cur_exp->ReadFile(full_item, 0, tik_file_size, tik_read_buf);
                            const auto rc = es::ImportTicket(tik_read_buf, tik_file_size, es::CommonCertificateData, es::CommonCertificateSize);
                            fs::DeleteWorkBuffer(tik_read_buf);
                            if(R_FAILED(rc)) {
                                HandleResult(rc, cfg::Strings.GetString(103));
                            }
                        }
                        break;
                    }
                }
            }
            /*
            else if(ext == "nxtheme") {
                switch(option_1) {
                    case 0: {
                        std::string nxthemes_nro_path = "sdmc:/switch/nxthemes_installer/nxthemesinstaller.nro";
                        auto sd_exp = fs::GetSdCardExplorer();
                        if(!sd_exp->IsFile(nxthemes_nro_path)) {
                            g_MainApplication->CreateShowDialog(cfg::Strings.GetString(104), cfg::Strings.GetString(105), { cfg::Strings.GetString(234) }, false);
                            return;
                        }
                        auto arg = nxthemes_nro_path + " installtheme=" + full_item;
                        size_t index = 0;
                        while(true) {
                            index = arg.find(" ", index);
                            if(index == std::string::npos) {
                                break;
                            }
                            arg.replace(index, 1, "(_)");
                        }
                        envSetNextLoad(nxthemes_nro_path.c_str(), arg.c_str());
                        g_MainApplication->CloseWithFadeOut();
                        return;
                    }
                }
            }
            */
            else if(ext == "nacp")  {
                switch(option_1) {
                    case 0: {
                        NacpStruct nacp_item = {};
                        const auto file_size = this->cur_exp->GetFileSize(full_item);
                        if(file_size < sizeof(NacpStruct)) {
                            g_MainApplication->ShowNotification(cfg::Strings.GetString(341));
                            return;
                        }
                        this->cur_exp->ReadFile(full_item, 0, sizeof(nacp_item), &nacp_item);

                        auto name = hos::FindNacpName(nacp_item);
                        if(name.empty()) {
                            name = cfg::Strings.GetString(106);
                        }
                        auto author = hos::FindNacpAuthor(nacp_item);
                        if(author.empty()) {
                            author = cfg::Strings.GetString(107);
                        }

                        auto msg = cfg::Strings.GetString(108) + "\n\n";
                        msg += cfg::Strings.GetString(91) + " " + name;
                        msg += "\n" + cfg::Strings.GetString(92) + " " + author;
                        msg += "\n" + cfg::Strings.GetString(109) + " " + nacp_item.display_version;

                        msg += "\n" + cfg::Strings.GetString(110) + " ";
                        switch(nacp_item.startup_user_account) {
                            case 0: {
                                msg += cfg::Strings.GetString(112);
                                break;
                            }
                            case 1: {
                                msg += cfg::Strings.GetString(111);
                                break;
                            }
                            case 2: {
                                msg += cfg::Strings.GetString(113);
                                break;
                            }
                            default: {
                                msg += cfg::Strings.GetString(114);
                                break;
                            }
                        }

                        msg += "\n" + cfg::Strings.GetString(115) + " ";
                        switch(nacp_item.screenshot) {
                            case 0: {
                                msg += cfg::Strings.GetString(111);
                                break;
                            }
                            case 1: {
                                msg += cfg::Strings.GetString(112);
                                break;
                            }
                            default: {
                                msg += cfg::Strings.GetString(114);
                                break;
                            }
                        }

                        msg += "\n" + cfg::Strings.GetString(116) + " ";
                        switch(nacp_item.video_capture) {
                            case 0: {
                                msg += cfg::Strings.GetString(112);
                                break;
                            }
                            case 1: {
                                msg += cfg::Strings.GetString(117);
                                break;
                            }
                            case 2: {
                                msg += cfg::Strings.GetString(111);
                                break;
                            }
                            default: {
                                msg += cfg::Strings.GetString(114);
                                break;
                            }
                        }

                        msg += "\n" + cfg::Strings.GetString(118) + " ";
                        switch(nacp_item.logo_type) {
                            case 0: {
                                msg += cfg::Strings.GetString(119);
                                break;
                            }
                            case 2: {
                                msg += cfg::Strings.GetString(120);
                                break;
                            }
                            default: {
                                msg += cfg::Strings.GetString(114);
                                break;
                            }
                        }

                        g_MainApplication->CreateShowDialog(cfg::Strings.GetString(58), msg, { cfg::Strings.GetString(234) }, false);
                        break;
                    }
                }
            }
            else if((ext == "jpg") || (ext == "jpeg")) {
                switch(option_1) {
                    case 0: {
                        if(!acc::HasSelectedUser()) {
                            g_MainApplication->ShowNotification(cfg::Strings.GetString(318));
                            return;
                        }

                        const auto option_2 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(121), cfg::Strings.GetString(122), { cfg::Strings.GetString(111), cfg::Strings.GetString(18) }, true);
                        if(option_2 < 0) {
                            return;
                        }

                        const auto file_size = this->cur_exp->GetFileSize(full_item);
                        auto icon_read_buf = fs::AllocateWorkBuffer(file_size);
                        this->cur_exp->ReadFile(full_item, 0, file_size, icon_read_buf);

                        const auto rc = acc::EditUserIcon(icon_read_buf, file_size);
                        if(R_SUCCEEDED(rc)) {
                            g_MainApplication->ShowNotification(cfg::Strings.GetString(123));
                        }
                        else {
                            HandleResult(rc, cfg::Strings.GetString(124));
                        }

                        fs::DeleteWorkBuffer(icon_read_buf);
                        break;
                    }
                }
            }
            else if(ext == "bin")  {
                switch(option_1) {
                    case 0: {
                        const auto option_2 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(125), cfg::Strings.GetString(126), { cfg::Strings.GetString(111), cfg::Strings.GetString(18) }, true);
                        if(option_2 < 0) {
                            return;
                        }
                        
                        const auto ok = hos::RebootWithPayload(full_item);
                        if(!ok) {
                            g_MainApplication->ShowNotification(cfg::Strings.GetString(460));
                        }
                        break;
                    }
                }
            }
            else if(!is_bin) {
                switch(option_1) {
                    case 0: {
                        g_MainApplication->LoadLayout(g_MainApplication->GetFileContentLayout());
                        g_MainApplication->GetFileContentLayout()->LoadFile(full_item, pres_full_item, this->cur_exp, false);
                        break;
                    }
                }
            }

            const auto view_option = option_count - 5;
            const auto copy_option = option_count - 4;
            const auto delete_option = option_count - 3;
            const auto rename_option = option_count - 2;
            if((option_1 == static_cast<s32>(view_option)) && (this->cur_exp->GetFileSize(full_item) > 0)) {
                g_MainApplication->LoadLayout(g_MainApplication->GetFileContentLayout());
                g_MainApplication->GetFileContentLayout()->LoadFile(full_item, pres_full_item, this->cur_exp, true);
            }
            else if(option_1 == static_cast<s32>(copy_option)) {
                UpdateClipboard(full_item);
            }
            else if(option_1 == static_cast<s32>(delete_option)) {
                this->PromptDeleteFile(full_item);
            }
            else if(option_1 == static_cast<s32>(rename_option)) {
                if(this->CheckWriteAccess()) {
                    const auto new_name = ShowKeyboard(cfg::Strings.GetString(130), item);
                    if(!new_name.empty()) {
                        if(new_name == item) {
                            return;
                        }
                        const auto new_path = this->cur_exp->FullPathFor(new_name);
                        if(this->cur_exp->IsFile(new_path) || this->cur_exp->IsDirectory(new_path)) {
                            HandleResult(rc::goldleaf::ResultEntryAlreadyPresent, cfg::Strings.GetString(254));
                        }
                        else {
                            this->cur_exp->RenameFile(full_item, new_path);
                            g_MainApplication->ShowNotification(cfg::Strings.GetString(133));
                            this->UpdateElements(this->browse_menu->GetSelectedIndex());
                        }
                    }
                }
            }
        }
    }

    void PartitionBrowserLayout::fsItems_Y(const std::string &item) {
        const auto full_item = this->cur_exp->FullPathFor(item);
        const auto pres_full_item = this->cur_exp->FullPresentablePathFor(item);

        const auto ipc_full_item = this->cur_exp->RemoveMountName(full_item);
        amssu::UpdateInformation update_info = {};
        auto rc = amssu::GetUpdateInformation(ipc_full_item.c_str(), &update_info);
        const auto is_valid_update = R_SUCCEEDED(rc);

        if(this->cur_exp->IsDirectory(full_item)) {
            const auto files = this->cur_exp->GetFiles(full_item);
            std::vector<std::string> nsps;
            for(const auto &file: files) {
                if(g_Settings.ignore_hidden_files && IsHiddenContent(file)) {
                    continue;
                }

                const auto path = full_item + "/" + file;
                const auto ext = LowerCaseString(fs::GetExtension(path));
                if(ext == "nsp") {
                    nsps.push_back(file);
                }
            }
            std::vector<std::string> extra_options = { cfg::Strings.GetString(281) };
            if(!nsps.empty()) {
                extra_options.push_back(cfg::Strings.GetString(282));
            }
            extra_options.push_back(cfg::Strings.GetString(18));

            auto msg = cfg::Strings.GetString(134);
            msg += "\n\n" + cfg::Strings.GetString(237) + " " + fs::FormatSize(this->cur_exp->GetDirectorySize(full_item));
            const auto option_1 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(135), msg, { cfg::Strings.GetString(415), cfg::Strings.GetString(73), cfg::Strings.GetString(74), cfg::Strings.GetString(75), cfg::Strings.GetString(280), cfg::Strings.GetString(18) }, true);
            switch(option_1) {
                case 0: {
                    if(is_valid_update) {
                        std::string update_msg = "";
                        const auto version_str = std::to_string((update_info.version >> 26) & 0x1F) + "." + std::to_string((update_info.version >> 20) & 0x1F) + "." + std::to_string((update_info.version >> 16) & 0xF);
                        update_msg += " - " + cfg::Strings.GetString(417) + " " + version_str + "\n\n";
                        update_msg += " - " + cfg::Strings.GetString(418) + " " + (update_info.exfat_supported ? cfg::Strings.GetString(111) : cfg::Strings.GetString(112));
                        
                        const auto option_2 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(416), update_msg, { cfg::Strings.GetString(419), cfg::Strings.GetString(18) }, true);
                        if(option_2 == 0) {
                            amssu::UpdateValidationInfo update_validation_info = {};
                            rc = amssu::ValidateUpdate(ipc_full_item.c_str(), &update_validation_info);
                            if(R_SUCCEEDED(rc) && R_SUCCEEDED(update_validation_info.result) && R_SUCCEEDED(update_validation_info.exfat_result)) {
                                const auto option_3 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(420), cfg::Strings.GetString(421), { cfg::Strings.GetString(423), cfg::Strings.GetString(18) }, true);
                                if(option_3 == 0) {
                                    const auto option_4 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(424), cfg::Strings.GetString(425), { cfg::Strings.GetString(111), cfg::Strings.GetString(112), cfg::Strings.GetString(18) }, true);
                                    if((option_4 == 0) || (option_4 == 1)) {
                                        const auto with_exfat = option_4 == 0;
                                        SetSysFirmwareVersion fw_ver = {};
                                        setsysGetFirmwareVersion(&fw_ver);
                                        g_MainApplication->LoadMenuData(cfg::Strings.GetString(424), "Update", std::string(fw_ver.display_version) + " → " + version_str + "...");
                                        g_MainApplication->LoadLayout(g_MainApplication->GetUpdateInstallLayout());
                                        g_MainApplication->GetUpdateInstallLayout()->InstallUpdate(ipc_full_item, with_exfat);
                                    }
                                }
                            }
                            else {
                                auto failed_rc = rc;
                                if(R_FAILED(update_validation_info.result)) {
                                    failed_rc = update_validation_info.result;
                                }
                                if(R_FAILED(update_validation_info.exfat_result)) {
                                    failed_rc = update_validation_info.exfat_result;
                                }
                                HandleResult(failed_rc, cfg::Strings.GetString(422));
                            }
                        }
                    }
                    else {
                        g_MainApplication->ShowNotification(cfg::Strings.GetString(433));
                    }
                    break;
                }
                case 1: {
                    UpdateClipboard(full_item);
                    break;
                }
                case 2: {
                    if(this->CheckWriteAccess()) {
                        const auto option_2 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(325), cfg::Strings.GetString(326), { cfg::Strings.GetString(111), cfg::Strings.GetString(18) }, true);
                        if(option_2 < 0) {
                            return;
                        }
                        this->cur_exp->DeleteDirectory(full_item);
                        g_MainApplication->ShowNotification(cfg::Strings.GetString(327));
                        this->UpdateElements();
                    }
                    break;
                }
                case 3: {
                    if(this->CheckWriteAccess())  {
                        const auto new_name = ShowKeyboard(cfg::Strings.GetString(238), item);
                        if(!new_name.empty()) {
                            if(new_name == item) {
                                // TODO: special handling?
                                return;
                            }
                            const auto new_path = this->cur_exp->FullPathFor(new_name);
                            if(this->cur_exp->IsFile(new_path) || this->cur_exp->IsDirectory(new_path)) {
                                HandleResult(rc::goldleaf::ResultEntryAlreadyPresent, cfg::Strings.GetString(254));
                            }
                            else {
                                this->cur_exp->RenameDirectory(full_item, new_path);
                                g_MainApplication->ShowNotification(cfg::Strings.GetString(139));
                                this->UpdateElements();
                            }
                        }
                    }
                    break;
                }
                case 4: {
                    const auto option_2 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(280), cfg::Strings.GetString(134), extra_options, true);
                    switch(option_2) {
                        case 0: {
                            this->cur_exp->SetArchiveBit(full_item);
                            this->UpdateElements(this->browse_menu->GetSelectedIndex());
                            g_MainApplication->ShowNotification(cfg::Strings.GetString(303));
                            break;
                        }
                        case 1: {
                            const auto option_3 = g_MainApplication->CreateShowDialog(cfg::Strings.GetString(77), cfg::Strings.GetString(78), { cfg::Strings.GetString(19), cfg::Strings.GetString(79), cfg::Strings.GetString(18) }, true);
                            if(option_3 < 0) {
                                return;
                            }
                            const auto dst = (option_3 == 0) ? NcmStorageId_SdCard : NcmStorageId_BuiltInUser;
                            for(const auto &nsp_name: nsps) {
                                const auto nsp_path = full_item + "/" + nsp_name;
                                const auto pres_nsp_path = pres_full_item + "/" + nsp_name;
                                
                                const auto file_size = this->cur_exp->GetFileSize(nsp_path);
                                const auto free_space = fs::GetFreeSpaceForPartition(static_cast<fs::Partition>(dst));
                                if(free_space < file_size) {
                                    HandleResult(rc::goldleaf::ResultNotEnoughSize, cfg::Strings.GetString(251));
                                    return;
                                }
                                g_MainApplication->LoadMenuHead(cfg::Strings.GetString(145) + " " + pres_nsp_path);
                                g_MainApplication->LoadLayout(g_MainApplication->GetInstallLayout());
                                g_MainApplication->GetInstallLayout()->StartInstall(nsp_path, this->cur_exp, dst, true);
                                g_MainApplication->LoadLayout(g_MainApplication->GetBrowserLayout());
                            }
                            this->ResetMenuHead();
                            break;
                        }
                    }
                    break;
                }
            }
        }
    }

}